/*******************************************************************************
 * Copyright 2022 ModalAI Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * 4. The Software is used solely in conjunction with devices provided by
 *    ModalAI Inc.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/


#ifndef MODALCV_PC_H
#define MODALCV_PC_H


#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>


void mcv_pc_en_debug(int en);
void mcv_pc_en_timing(int en);



/**
 * Used to store an obstacle point cloud after it's been downsampled.
 * these are then kept in an ob_pc_ringbuf_t to be rotated and itterated through
 */
typedef struct mcv_pc_t{
	int initialized;	///< initialization flag
	int n;				///< number of XYZ points
	int64_t ts_ns;		///< timestamp when sensor data was collected
	float* d;			///< pointer to dynamically allocated data
	int sensor_type;	//< optional,  DFS/TOF
	char frame[64];		//< optional, name of coordinate frame the point cloud is in, usually "stereo" or "tof"
} mcv_pc_t;

#define MCV_PC_INITIALIZER {\
	.initialized = 0,\
	.n = 0,\
	.ts_ns = 0,\
	.d = NULL,\
	.sensor_type = 0,\
	.frame = {0}\
}

int mcv_pc_alloc(mcv_pc_t* pc, int n);
int mcv_pc_free(mcv_pc_t* pc);
int mcv_pc_print(mcv_pc_t pc);
mcv_pc_t mcv_pc_empty(void);


/**
 * @brief      downsample a high-density point cloud
 *
 *             creates and allocates a new obstable point cloud from a list of
 *             raw points. It downsamples by binning into a 3D grid and filters
 *             out stray floaters tunable by the threshold argument.
 *
 *             This is intended for point clouds from DFS or TOF sensors where Z
 *             points out the sensor. As such, it throws out any points with a Z
 *             value <=0 or >max_depth which discards the "zero" and "inf"
 *             points that DFS and TOF generate.
 *
 *             Cell size is the resolution of the 3D grid, a larger number will
 *             result in a more sparse output point cloud. Recommend something
 *             around 0.15
 *
 *             cells in the grid with fewer than "threshold" occupied cells in
 *             the surrounding 3x3 block will be discarded. This includes the
 *             cell itself, so threshold cannot be less than 1 or greater than
 *             27. Setting threshold to 1 disables this check, since the
 *             occupied cell itself counts in the check. The intention of this
 *             filter is to remove stray floating points which are likely noise.
 *
 *             confidence checking is optional. Set conf to NULL to ignore.
 *
 * @param[in]  n            number of input points
 * @param      in           array of XYZ points
 * @param[in]  conf         confidence from 0 (min) to 255 (very confident)
 * @param[in]  conf_cutoff  The conf cutoff, points with conf less than this
 *                          will be discarded
 * @param[in]  max_depth    The maximum depth in +Z direction allowed
 * @param[in]  cell_size    The cell size
 * @param[in]  threshold    The threshold
 * @param      out          pointer to pc struct to store the data
 *
 * @return     0 on success, -1 on failure
 */
int mcv_pc_downsample(int n, const float* in, const uint8_t* conf, uint8_t conf_cutoff,\
				float max_depth, float cell_size, int threshold, mcv_pc_t* out);


// generic point cloud downsampler, does not confidence checking, or box filtering
// unlike mcv_pc_downsample which only allows positive z and is meant for depth sensor measurements,
// this is meant for more generic point clouds
int mcv_pc_downsample_generic_no_filter(int n, const float* in, float cell_size, mcv_pc_t* out);




// configuration params for generating 3d point cloud from disparity image
// use the skip_lines feature to reduce the point cloud size. set to 0 to output
// every point. 1 will skip every other row and column resulting in 1/4 the points
// set to 2 to get 1/9 the points, etc
typedef struct mcv_pc_disp_config_t{
	int width;
	int height;
	float f;
	float cx;
	float cy;
	float baseline_m;
	int skip_lines;
	uint8_t min_disparity; // points with disparity less than this will be discarded
} mcv_pc_disp_config_t;

#define MCV_PC_DISP_CONFIG_INITIALIZER {\
	.width = 0,\
	.height = 0,\
	.f = 0,\
	.cx = 0,\
	.cy = 0,\
	.baseline_m = 0,\
	.skip_lines = 0,\
	.min_disparity = 0 \
}

// lookup table used by mcv_pc_from_disparity. Set it up with mcv_pc_disp_table_init()
typedef struct mcv_pc_disp_table_t{
	int initialized;
	mcv_pc_disp_config_t config;
	float fb_over_d[256];
	float b_over_d[256];
} mcv_pc_disp_table_t;

#define MCV_PC_DISP_TABLE_INITIALIZER {\
	.initialized = 0,\
	.config = MCV_PC_DISP_CONFIG_INITIALIZER,\
	.fb_over_d = {0},\
	.b_over_d = {0} \
}


// initialize a lookup table used by mcv_pc_from_disparity before generating a
// point cloud. Only needs to be done once for a given config.
int mcv_pc_disp_table_init(mcv_pc_disp_config_t config, mcv_pc_disp_table_t* table);

/**
 * @brief      create a 3d point cloud from a disparity image
 *
 * @param[in]  <unnamed>  { parameter_description }
 *
 * @return     { description_of_the_return_value }
 */
int mcv_pc_from_disparity(uint8_t* disp_img, mcv_pc_disp_table_t* table, mcv_pc_t* pc);


#ifdef __cplusplus
}
#endif

#endif // end #define MODALCV_PC_H
